using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.Diagnostics;
using TelnetExpect.Ntlm;

namespace TelnetExpect
{
    /// <summary>
    /// Enables automatic login to a server with NTLM (TNAP) authentication.
    /// </summary>
    public abstract class AuthenticationNtlmTelnetHandler : TelnetOptionHandler
    {
        const bool IncludeSpnTarget = true;
        //
        const string SpnServiceClass = "telnet";
        //
        bool? m_promptOrForce;
        MySspi m_sspi;
        string m_targetHostName;

        protected AuthenticationNtlmTelnetHandler(bool? promptOrForce)
        {
            m_promptOrForce = promptOrForce;
        }

        protected AuthenticationNtlmTelnetHandler()
            : this(null)
        {
        }

        public void SetTargetHostName(string hostName)
        {
            if (string.IsNullOrEmpty(hostName))
                throw new ArgumentNullException("hostName");
            m_targetHostName = hostName;
        }

        public void ClearTargetHostName()
        {
            m_targetHostName = null;
        }

        //----
        bool Allow()
        {
            if (m_promptOrForce != null) {
                return m_promptOrForce.Value;
            }
            //
            bool useTnap = Prompt("Do you want to send your NTLM credentials to the remote device?");
            return useTnap;
        }

        protected abstract bool Prompt(string message);

        /// <exclude/>
        protected internal override void Init()
        {
            base.Init();
            // We (optionally) prompt the user only if/when the server offers us TNAP Auth.
            Telnet.SetLocalMode(TelnetOption.AuthenticationOption, true);
        }

        /// <exclude/>
        protected override void HandleSubNegotiation(byte[] data)
        {
            TelnetOption id = (TelnetOption)data[0];
            if (id == TelnetOption.AuthenticationOption) {
                DumpAuthSubNeg(data, "Rcvd");
                Debug.Assert(Telnet.GetLocalMode(TelnetOption.AuthenticationOption), "NOT Option Allowed locally");
                Debug.Assert(Telnet.GetLocalState(TelnetOption.AuthenticationOption), "NOT Option Enabled locally");
                if (data.Length < 2) {
                    //Need at least option code and command code.
                    Error();
                    return;
                }
                AuthCmd ac = (AuthCmd)data[1];
                switch (ac) {
                    case AuthCmd.SEND:
                        HandleSendCommand(data);
                        return;
                    case AuthCmd.REPLY:
                        HandleReplyCommand(data);
                        return;
                    case AuthCmd.IS:    // These two not expected from server!
                    case AuthCmd.NAME:
                    default:            // Shouldn't be any others.
                        Error();
                        break;
                }
                //
            }
        }

        private void HandleReplyCommand(byte[] data)
        {
            if (data.Length < 4) {
                Error();
                return;
            }
            Debug.Assert(data.Length >= 4);
            Debug.Assert((TelnetOption)data[0] == TelnetOption.AuthenticationOption);
            Debug.Assert((AuthCmd)data[1] == AuthCmd.REPLY);
            AuthType at = (AuthType)data[2];
            AuthModifiers am = (AuthModifiers)data[3];
            if (data.Length < 5) {
                Error();
                return;
            }
            NtlmCmd cmd = (NtlmCmd)data[4];
            if (cmd == NtlmCmd.Reject) {
                // Failed authentication; fallback to username/password authentication.
                return;
            }
            if (cmd == NtlmCmd.Accept) {
                // Success authentication.
                return;
            }
            Debug.Assert(cmd == NtlmCmd.Challenge, string.Format(
                "Should be Challenge packet (is {0}==0x{1:X2}).",cmd, (int)cmd));
            if (data.Length < 9) {
                Error();
                return;
            }
            int offset = 5;
            int len = BitConverter.ToInt32(data, offset);
            Debug.Assert(len >=0, "Wrapped UInt32!?!");
            Debug.Assert((len & 0xff) == data[offset], "Not read as litte-endian?!!");
            offset += 4;
            int bufType = BitConverter.ToInt32(data, offset);
            Debug.Assert(bufType == 2, "Unexpected bufType");
            offset += 4;
            Debug.Assert(len == data.Length - offset, "Length prefix value not correct??");
            //
            byte[] ntlmIn = new byte[len];
            Array.Copy(data, offset, ntlmIn, 0, ntlmIn.Length);
            const UInt64 ExpectedMagicNumber = 0x005053534d4c544e; // "NTLMSSP\0";
            UInt64 magicNumberX = BitConverter.ToUInt64(data, offset);
            Debug.Assert(magicNumberX == ExpectedMagicNumber, "Not NTLMSSP?");
            UInt64 magicNumber = BitConverter.ToUInt64(ntlmIn, 0);
            Debug.Assert(magicNumber == ExpectedMagicNumber, "Not NTLMSSP?");
            //
            //byte[] dummyResponse = CreateDummyAuthPacketWithTnapHeader();
            //byte[] reply = CreateTnapSubNegResponse(AuthCmd.IS, dummyResponse, data[2], data[3]);
            byte[] authData = CreateAuthPacket(ntlmIn);
            authData = WrapInTnapHeader(NtlmCmd.Authenticate, authData);
            byte[] reply = CreateTnapSubNegResponse(AuthCmd.IS, authData, data[2], data[3]);
            DumpAuthSubNeg(reply, "Send");
            AddResponseSubNegotiation(reply);
        }

        private static byte[] CreateDummyAuthPacketWithTnapHeader()
        {
            byte[] dummySspResponse = new byte[0xAC];
            byte[] tnapPrefix = {
                // TNAP
                0x02, // TNAP_AUTHENTICATE
                0xAC, 0x00, 0x00, 0x00, // Size
                0x02, 0x00, 0x00, 0x00, // Buffer Type
                // NTLM
            };
            byte[] dummyResponse = new byte[tnapPrefix.Length + dummySspResponse.Length];
            tnapPrefix.CopyTo(dummyResponse, 0);
            dummySspResponse.CopyTo(dummyResponse, tnapPrefix.Length);
            return dummyResponse;
        }

        private void HandleSendCommand(byte[] data)
        {
            List<AuthTypePair> list = ParseAuthSendCommand(data);
            DumpAuthTypePairList(list);
            AuthTypePair chosen = null;
            foreach (AuthTypePair cur in list) {
                if (cur.AuthType == AuthType.NTLM
                        && cur.Modifers == AuthModifiers.None) {
                    chosen = cur;
                    break;
                }
                if (cur.AuthType == AuthType.NTLM) {
                    //DEBUG-> NTLM but wrong modifiers
                }
            }//for
            if (chosen == null || !Allow()) {
                byte[] nullSelection = {
                        (byte)TelnetOption.AuthenticationOption,
                        (byte)AuthCmd.IS,
                        (byte)AuthType.NULL,
                        (byte)AuthModifiers.None,
                    };
                DumpAuthSubNeg(nullSelection, "Send");
                AddResponseSubNegotiation(nullSelection);
                return;
            }
            // The one!
            byte[] authData = CreateNegoPacket();
            authData = WrapInTnapHeader(NtlmCmd.Negotiate, authData);
            byte[] reply = CreateTnapSubNegResponse(AuthCmd.IS, authData, data[2], data[3]);
            DumpAuthSubNeg(reply, "Send");
            AddResponseSubNegotiation(reply);
            return;
        }

        private void DumpAuthTypePairList(List<AuthTypePair> list)
        {
            const string Sepa = "; ";
            StringBuilder bldr = new StringBuilder();
            foreach (AuthTypePair cur in list) {
                bldr.AppendFormat(System.Globalization.CultureInfo.InvariantCulture,
                    "'{0}' '{1}'", cur.AuthType, cur.Modifers);
                bldr.Append(Sepa);
            }
            if (bldr.Length == 0) {
                bldr.Append("<empty list>");
            } else {
                bldr.Length -= Sepa.Length;
            }
            TraceWriteLine("authentication-type-pair-list: [ {0} ].", bldr.ToString());
        }

        private static byte[] CreateTnapSubNegResponse(AuthCmd cmd, byte[] authData, byte data2, byte data3)
        {
            byte[] reply = new byte[4 + authData.Length];
            reply[0] = (byte)TelnetOption.AuthenticationOption;
            reply[1] = (byte)cmd;
            reply[2] = (byte)AuthType.NTLM;
            reply[3] = (byte)AuthModifiers.None;
            Debug.Assert(reply[2] == data2);
            Debug.Assert(reply[3] == data3);
            //Array.Copy(data, 2, reply, 2, data.Length - 2);
            Array.Copy(authData, 0, reply, 2 + 2, authData.Length);
            return reply;
        }

        private List<AuthTypePair> ParseAuthSendCommand(byte[] data)
        {
            Debug.Assert(data.Length >= 2, "need two (option code and command) bytes -- should have been checked by caller!");
            Debug.Assert((TelnetOption)data[0] == TelnetOption.AuthenticationOption);
            Debug.Assert((AuthCmd)data[1] == AuthCmd.SEND);
            int offset = 2;
            List<AuthTypePair> list = new List<AuthTypePair>();
            TraceWriteLineIf(data.Length < 4, "*Empty* authentication-type-pair-list received!");
            while (true) {
                // Check this first, to allow *empty* auth-type-pair-list!!!
                if (data.Length - offset < 2) {
                    TraceWriteLineIf(data.Length - offset != 0, "Odd length authentication-type-pair-list received!");
                    break;
                }
                byte authType = data[offset + 0];
                byte modifiers = data[offset + 1];
                list.Add(new AuthTypePair(authType, modifiers));
                offset += 2;
            }//while
            return list;
        }

        class AuthTypePair
        {
            readonly AuthType m_authType;
            readonly AuthModifiers m_modifiers;

            public AuthTypePair(byte authType, byte modifiers)
                : this(checked((AuthType)authType), checked((AuthModifiers)modifiers))
            {
            }

            public AuthTypePair(AuthType authType, AuthModifiers modifiers)
            {
                m_authType = authType;
                m_modifiers = modifiers;
            }

            public AuthType AuthType
            {
                get { return m_authType; }
            }

            public AuthModifiers Modifers
            {
                get { return m_modifiers; }
            }
        }//class

        byte[] CreateNegoPacket()
        {
            bool done;
            Debug.Assert(m_sspi == null, "This method should be called first!");
            if (IncludeSpnTarget && m_targetHostName != null) {
                string spn = MySspi.CreateSpn(SpnServiceClass, m_targetHostName, TelnetBaseStream.DefaultPort, TelnetBaseStream.DefaultPort);
                m_sspi = new MySspi(spn);
            } else {
                m_sspi = new MySspi();
            }
            byte[] send = m_sspi.DoStep(null, out done);
            return send;
        }

        byte[] CreateDummyNegoPacketWithTnapHeader()
        {
            byte[] template = {
                // TNAP
                0x00, // TNAP_NEGO
                0x28, 0x00, 0x00, 0x00, // Size = 0x28 = 42
                0x02, 0x00, 0x00, 0x00, // Buffer Type
                // NTLM
                0x4E, 0x54, 0x4C, 0x4D, 0x53, 0x53, 0x50, 0x00, //NTLMSSP.
                0x01, 0x00, 0x00, 0x00, 0x97, 0x82, 0x08, 0xE2, //....??.
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //........
                0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, //........
                0x05, 0x01, 0x28, 0x0A, 0x00, 0x00, 0x00, 0x0F  //..(..... 
            };
            return (byte[])template.Clone();
        }

        byte[] CreateAuthPacket(byte[] challengeData)
        {
            bool done;
            byte[] send = m_sspi.DoStep(challengeData, out done);
            return send;
        }

        byte[] WrapInTnapHeader(NtlmCmd cmd, byte[] ntlmssp)
        {
            const int headerLen = 9;
            byte[] all = new byte[headerLen + ntlmssp.Length];
            ntlmssp.CopyTo(all, headerLen);
            all[0] = checked((byte)cmd);
            BitConverter.GetBytes(ntlmssp.Length).CopyTo(all, 1);
            const int BufferType = 2;
            BitConverter.GetBytes(BufferType).CopyTo(all, 5);
            return all;
        }

        private void DumpAuthSubNeg(byte[] data, string prefix)
        {
            AuthCmd? ac = null;
            AuthType? at = null;
            AuthModifiers? am = null;
            NtlmCmd? cmd = null;
            if (data.Length >= 2) {
                ac = (AuthCmd)data[1];
                if (data.Length >= 3) {
                    at = (AuthType)data[2];
                    if (data.Length >= 4) {
                        am = (AuthModifiers)data[3];
                        if (data.Length >= 5) {
                            cmd = (NtlmCmd)data[4];
                        }
                    }
                }
            }
            TraceWriteLine("{0} OptionAuth: Cmd: {1}, Type: {2}, Modifier: {3}, leftover: {4}, (NtlmCmd: {5})",
                prefix, ac, at, am, data.Length - 4, cmd);
        }

        private void Error()
        {
            throw new Exception("The method or operation is not implemented.");
        }

        enum AuthCmd
        {
            IS = 0,
            SEND = 1,
            REPLY = 2,
            NAME = 3
        }

        enum AuthType
        {
            NULL = 0,
            KERBEROS_V4 = 1,
            KERBEROS_V5 = 2,
            SPX/* * */ = 3,
            MINK/* * */= 4,
            SRP = 5,
            RSA/* *[also used by SRA*]*/= 6,
            SSL /* * */= 7,
            //[unassigned]8,
            //[unassigned]9,
            LOKI/* * */ = 10,
            SSA/* * */ = 11,
            KEA_SJ = 12,
            KEA_SJ_INTEG = 13,
            DSS = 14,
            NTLM/* * */ = 15
        }

        [Flags]
        enum AuthModifiers
        {
            None = 0,
            //
            //WhoClientToServer = 0,
            WhoServerToClient = 1,
            //
            // HowOneWay = 0,
            HowMutual = 2,
            //
            CredForwarding = 8,
            //
            //EncryptOff = 0,
            EncryptUsingTelopt = 4,
            EncryptAfterExch = 0x10,
            EncryptReserved = 0x14,
        }

        static class AuthModifiers_Masks
        {
            public const AuthModifiers AuthWho = (AuthModifiers)0x01;
            public const AuthModifiers AuthHow = (AuthModifiers)0x02;
            public const AuthModifiers Encrypt = (AuthModifiers)0x14;
            public const AuthModifiers CredForwarding = (AuthModifiers)0x08;
        }

        enum NtlmCmd
        {
            Negotiate = 0x00,       // Client
            Challenge = 0x01,       // server
            Authenticate = 0x02,    // Client
            Accept = 0x03,          // server
            Reject = 0x04,          // server
        }

    }
}
